/*	$Id: pom.c 2007/08/27 12:06:17 mallet $ */

/*
 * Copyright (c) 2007 Anthony Mallet                  --  Thu Aug 23 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <err.h>
#include <zlib.h>

#include "libpom/lib.h"

#define TOLERANCE	1e-4
#define ITERATIONS	1000
#define SRAND		0 /*time(NULL)*/

void	randompos(LibPomMEPos *pos, int date);
void	readpos(LibPomMEPos *pos, int date, gzFile *input);
void	savepos(T3D *pos, gzFile *regen);
int	gzscanfd(gzFile *f, double *d);

int
main(int argc, const char *argv[])
{
  T3D tmp;
  LibPomMEPos pos;
  LibPomPos fuse;
  int me1, me2;
  int s, date;
  int i, iter;
  gzFile *regen, *input;

  if (argc > 2 && !strcmp(argv[2], "regen")) {
    input = NULL;
    regen = gzopen("fuse.data.gz", "w9");
    if (!regen) { warnx("cannot open file fuse.data"); exit(2); }
  } else if (argc > 2) {
    regen = NULL;
    input = gzopen(argv[2], "r");
    if (!input) { warnx("cannot open file %s", argv[1]); exit(2); }
  } else {
    warnx("usage: %s method regen|<file>", argv[0]); exit(2);
  }

  /* create two ME */
  if (!strcmp(argv[1], "kalman"))
    s = pomSetFusionMethod(POM_KALMAN);
  else
    s = pomSetFusionMethod(POM_MAX_CONF);
  if (s) { warnx("cannot init pom"); exit(2); }

  me1 = pomMECreate("test1", 1);
  if (me1 < 0) { warnx("cannot create ME"); exit(2); }
  me2 = pomMECreate("test2", 1);
  if (me2 < 0) { warnx("cannot create ME"); exit(2); }
  tmp = t3dUBryanIdentity;
  s = pomSetZero(&tmp);
  if (s) { warnx("cannot set zero"); exit(2); }

  s = pomSetRefME(me1);
  if (s) { warnx("cannot set ref ME"); exit(2); }

  /* feed positions */
  if (regen) srand48(SRAND);
  for(iter = 0; iter < ITERATIONS; iter++) {

    date = pomTick();

    if (input)
      /* read position */
      readpos(&pos, date, input);
    else
      /* make random positions */
      randompos(&pos, date);

    s = pomMEUpdate(me1, &pos);
    if (s) { warnx("cannot update ME1 at iteration %d", iter); exit(2); }
    if (regen) {
      gzprintf(regen, "%.20e\n", pos.confidence);
      savepos(&pos.pos.abs.mainToOrigin, regen);
    }

    if (input)
      /* read position */
      readpos(&pos, date, input);
    else
      /* make random positions */
      randompos(&pos, date);

    s = pomMEUpdate(me2, &pos);
    if (s) { warnx("cannot update ME2 at iteration %d", iter); exit(2); }
    if (regen) {
      gzprintf(regen, "%.20e\n", pos.confidence);
      savepos(&pos.pos.abs.mainToOrigin, regen);
    }

    /* fuse */
    s = pomFuse(&fuse);
    if (s) { warnx("cannot fuse at iteration %d", iter); exit(2); }

    if (input) {
      double cmp;

      t3dConvertTo(T3D_U_BRYAN, &fuse.mainToBase);
      for(i=0;i<6;i++) {
	if (gzscanfd(input, &cmp) != 1) exit(2);
	if (fabs(cmp - fuse.mainToBase.ubryan.bryan[i]) > TOLERANCE) {
	  warnx("bad fused position at iteration %d", iter); exit(2);
	}
      }
      for(i=0;i<21;i++) {
	if (gzscanfd(input, &cmp) != 1) exit(2);
	if (fabs(cmp - fuse.mainToBase.ubryan.v[i]) > TOLERANCE) {
	  warnx("bad fused position at iteration %d", iter); exit(2);
	}
      }

      t3dConvertTo(T3D_U_BRYAN, &fuse.mainToOrigin);
      for(i=0;i<6;i++) {
	if (gzscanfd(input, &cmp) != 1) exit(2);
	if (fabs(cmp - fuse.mainToOrigin.ubryan.bryan[i]) > TOLERANCE) {
	  warnx("bad fused position at iteration %d", iter); exit(2);
	}
      }
      for(i=0;i<21;i++) {
	if (gzscanfd(input, &cmp) != 1) exit(2);
	if (fabs(cmp - fuse.mainToOrigin.ubryan.v[i]) > TOLERANCE) {
	  warnx("bad fused position at iteration %d", iter); exit(2);
	}
      }
    } else {
      savepos(&fuse.mainToBase, regen);
      savepos(&fuse.mainToOrigin, regen);
    }
  }

  warnx("pomFuse succeeded on %d random positions and two estimators.", iter);
  return 0;
}


void
randompos(LibPomMEPos *pos, int date)
{
#define INDEX_SMATELT(i, j)   \
   ((i)<(j) ? ((j)*((j)+1))/2 + (i) : ((i)*((i)+1))/2 +(j))
#define SMATELT(mat, i, j, n)	(*((mat) + INDEX_SMATELT(i,j)))

  int i,j,k;
  double val, v[21];

  pos->class = POM_ABSOLUTE_ME;
  pos->posType = POM_ABSOLUTE_ME;
  pos->confidence = drand48();

  pos->pos.abs.date = date;
  pos->pos.abs.mainToOrigin = t3dUBryanIdentity;
  pos->pos.abs.mainToOrigin.flags = T3D_ALLOW_CONVERSION;

  /* position */
  for(i=0; i<6; i++)
    pos->pos.abs.mainToOrigin.ubryan.bryan[i] = drand48();

  /* covariances (make a symmetric positive definite matrix) */
  for(i=0; i<21; i++) v[i] = 20*drand48();
  for (i=0; i<6; i++) for (j=i; j<6; j++) {
    val = 0.;
    for (k=0; k<6; k++)
      val += SMATELT(v, i, k, 6) * SMATELT(v, k, j, 6);
    SMATELT(pos->pos.abs.mainToOrigin.ubryan.v, i, j, 6) = val;
  }

  /* velocity */
  pos->velocity.date = date;
  pos->velocity.VLocal = t3dUMatrixIdentity;
  pos->velocity.VLocal.flags = T3D_ALLOW_CONVERSION;
}


void
readpos(LibPomMEPos *pos, int date, gzFile *input)
{
  int i;

  pos->class = POM_ABSOLUTE_ME;
  pos->posType = POM_ABSOLUTE_ME;
  if (gzscanfd(input, &pos->confidence) != 1) exit(2);

  /* position */
  pos->pos.abs.date = date;
  t3dInit(&pos->pos.abs.mainToOrigin, T3D_U_BRYAN, T3D_ALLOW_CONVERSION);
  for(i=0;i<6;i++)
    if (gzscanfd(input, &pos->pos.abs.mainToOrigin.ubryan.bryan[i]) != 1)
      exit(2);

  for(i=0;i<21;i++)
    if (gzscanfd(input, &pos->pos.abs.mainToOrigin.ubryan.v[i]) != 1)
      exit(2);

  /* velocity */
  pos->velocity.date = date;
  pos->velocity.VLocal = t3dUMatrixIdentity;
  pos->velocity.VLocal.flags = T3D_ALLOW_CONVERSION;
}


void
savepos(T3D *t3d, gzFile *regen)
{
  int i;

  t3dConvertTo(T3D_U_BRYAN, t3d);
  for(i=0;i<6;i++)
    gzprintf(regen, "%.20e ", t3d->ubryan.bryan[i]);
  gzprintf(regen, "\n");

  for(i=0;i<21;i++)
    gzprintf(regen, "%.20e ", t3d->ubryan.v[i]);
  gzprintf(regen, "\n");
}

int
gzscanfd(gzFile *f, double *d)
{
  char b[64], *p;

  for(p=b; isspace(*p = gzgetc(f));)		/* empty body */;
  for(p++; !isspace(*p = gzgetc(f)); p++)	/* empty body */;
  *p = '\0';

  return sscanf(b, "%lg", d);
}
