/*	$Id: matrix.c 2007/08/24 14:56:30 mallet $ */

/*
 * Copyright (c) 1990,2004 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

/* ATTENTION (for LAAS users):
 *
 * This file contains some functions that are in libmatrices from
 * calife. All the dm_* functions have been prefixed by pom_dm_*.
 *
 * These functions should probably be replaced by BLAS routines.
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>

#include "libpom/lib.h"
#include "pomInt.h"

/* ----------------------------------------------------------------------
 *
 * Operations on matrices.
 *
 * --- Full matrices:
 *
 *  [ 1 2 3 ]
 *  [ 4 5 6 ]
 *  [ 7 8 9 ]
 *
 *  is stored as [1 2 3 4 5 6 7 8 9].
 *
 * --- Symmetric matrices:
 *
 *  [ 1 2 4 ]
 *  [ 2 3 5 ]
 *  [ 4 5 6 ]
 *
 *  is stored as [1 2 3 4 5 6].
 */


/* --- local data ---------------------------------------------------- */

#define MATELT(mat, i, j, c)	(*((mat) + (c) * (i) + (j)) )

#define INDEX_SMATELT(i, j)   \
   ((i)<(j) ? ((j)*((j)+1))/2 + (i) : ((i)*((i)+1))/2 +(j))
#define SMATELT(mat, i, j, n)	(*((mat) + INDEX_SMATELT(i,j)))


/* --- pom_dm_fill ------------------------------------------------------
 *
 * fill matrix m by values
 *
 *	with: m full		lxc
 *	      val1, ...		double
 *
 * Example : t3d_dm_fill(mat, 2, 3, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0)
 *
 *     m = [ 1.0 2.0 3.0 ]
 *	   [ 4.0 5.0 6.0 ]
 */

int
pom_dm_fill(double *mat, int l, int c, ...)
{
   va_list ap;
   int i;

   va_start(ap, c);

   for (i = 0; i < l*c; i++)
      *mat++ = va_arg (ap, double);

   va_end (ap);
   return 0;
}


/* --- pom_dms_mulscal_6x6 -------------------------------------------
 *
 * produc     m = scal * m1
 *
 *	with : m symetric	6x6
 *
 * Result is symetric, 6x6.
 */

int
pom_dms_mulscal_6x6(double *m, double scal)
{
   int k;

   for (k = 0; k < 21; k++)
      *m++ *= scal;

   return 0;
}


/* --- pom_dm_mult ------------------------------------------------------
 *
 * product     m1 * m2
 *
 *	with : m1 full		l1xc1
 *	       m2 full		l2xc2
 *
 * Result is full, l1xc2.
 */

int
pom_dm_mult(const double *m1, const double *m2, double *m,
	    int l1, int c1, int l2, int c2)
{
   int i, j, k;
   double val;

   if (c1 != l2) {
     fprintf(stderr, "invalid call of t3d_dm_mult\n");
     return 1;
   }
   if (m == m1 || m == m2) {
     fprintf(stderr, "invalid call of t3d_dm_mult\n");
     return 1;
   }

   for (i = 0; i < l1; i++)
      for (j = 0; j < c2; j++) {
	 val = 0.;
	 for (k = 0; k < c1; k++)
	    val += MATELT(m1, i, k, c1) * MATELT(m2, k, j, c2);
	 *m++ = val;
      }

   return 0;
}


/* --- pom_dmsf_mult -------------------------------------------------
 *
 * product     m1 * m2
 *
 *	with : m1 full		l1xc1
 *	       m2 symetric	l2xc2
 *
 * Result is full, l1xc2.
 */

int
pom_dmsf_mult(double *m1, double *m2, double *m, int l1, int c1, int l2, int c2)
{
  int i,j,k;
  double val;

  if (c1 != l1 || c1 != l2 || m == m1 || m == m2) {
    fprintf(stderr, "invalid call of pom_dmsf_mult\n");
    return 1;
  }

  for (i = 0; i < l1; i++)
    for (j = 0; j < c2; j++) {
      val = 0.;
      for (k = 0; k < c1; k++)
	val += SMATELT (m1, i, k, c1) * MATELT (m2, k, j, c2);
      *m++ = val;
    }

  return 0;
}


/* --- pom_dmsf_multtr_6x6 -------------------------------------------
 *
 * product     m1 * m2_transposed
 *
 *	with : m1 full		6x6
 *	       m2 full		6x6
 *
 * Result is full, 6x6.
 */

int
pom_dmsf_multtr_6x6(double *m1, double *m2, double *m)
{
  int i,j,k;
  double val;

  if (m == m1) {
    fprintf(stderr, "invalid call of t3d_dm_mult\n");
    return 1;
  }

  for (i = 0; i < 6; i++)
    for (j = 0; j < 6; j++) {
      val = 0.;
      for (k = 0; k < 6; k++)
	val += SMATELT (m1, i, k, 6) * MATELT (m2, j, k, 6);
      *m++ = val;
    }

  return 0;
}


/* --- pom_dm_trans_3x3 -------------------------------------------------
 *
 * transpose     m
 *
 *	with : m full		3x3
 *
 * Result is full, 3x3.
 */

int
pom_dm_trans_3x3(double *m, double *r)
{
   r[0] = m[0]; r[1] = m[3]; r[2] = m[6];
   r[3] = m[1]; r[4] = m[4]; r[5] = m[7];
   r[6] = m[2]; r[7] = m[5]; r[8] = m[8];
   return 0;
}


/* --- pom_dm_trans_6x6 ----------------------------------------------
 *
 * transpose     m
 *
 *	with : m full		6x6
 *
 * Result is full, 6x6.
 */

int
pom_dm_trans_6x6(double *m, double *r)
{
  int i,j;

  for(i = 0; i < 6; i++) for(j = 0; j < 6; j++)
    r[i * 6 + j] = m[j * 6 + i];
   return 0;
}


/* --- pom_dm_plustr_6x6 ---------------------------------------------
 *
 * m = m1 + m1_transposee (m sym)
 */

int
pom_dm_plustr_6x6(double *m1, double *m)
{
  int i,j;

  for (i = 0; i < 6; i++)
    for (j = i; j < 6; j++)
      *(m + j*(j+1)/2 + i) = MATELT(m1, i,j, 6) + MATELT(m1, j,i, 6);

  return 0;
}


/* --- pom_dms_plus_6x6 ----------------------------------------------
 *
 * sum		m1 + m2
 *
 *	with:	m1 symmetric	6x6
 *		m2 symmetric	6x6
 *
 * Result is symmetric
 */

int
pom_dms_plus_6x6(double *m1, double *m2, double *m)
{
   int k;

   for (k = 0; k < 21; k++)
      *m++ = *m1++ + *m2++;
   return 0;
}


/* --- pom_dmfsf_jmult_6x6 -------------------------------------------
 *
 * product     m1 * m2 * m1_transposed
 *
 *	with : m1 full		6x6
 *	       m2 symmetric	6x6
 *
 * Result is always symmetric.
 *
 * t3d_dmfsf_jmult_6x6(m1, m2, m) ---> m = m1 * m2 * m1_transposed
 */

int
pom_dmfsf_jmult_6x6(double *m1, double *m2, double *m)
{
   double aux[6*6], val;
   double *temp;
   int i, j, k;

   if (m == m1 || m == m2) {
     fprintf(stderr, "invalid call of t3d_dmfsf_jmult_6x6\n");
     return 1;
   }

   temp = aux;
   for (i = 0; i < 6; i++)
      for (j = 0; j < 6; j++) {
	 val = 0.;
	 for (k = 0; k < 6; k++)
	    val += MATELT(m1, i, k, 6) * SMATELT(m2, k, j, 6);
	*temp++ = val;
      }

   temp = aux;
   for (j = 0; j < 6; j++)
      for (i = 0; i <= j; i++) {
	 val = 0.;
	 for (k = 0; k < 6; k++)
	    val += MATELT(temp, i, k, 6) * MATELT(m1, j, k, 6);
	 *m++ = val;
      }

   return 0;
}


/* --- pom_dms_inv_6x6 ----------------------------------------------- */

/* inversion m2 = m1_inverse 
 *
 *	with : m1 symmetric	6x6
 *
 * Result is symmetric 6x6
 */

int
pom_dms_inv_6x6(double *m1, double *m2)
{
  extern void dpotf2_(char *, int *, double *, int *, int *);
  int i,j,k;
  double sum;
  int r, e;
  double a[36], b[36];

  /* copy symmetric matrix in full matrix */
  for(i=0; i<6; i++) for(j=0; j<6; j++)
    MATELT(a, i, j, 6) = SMATELT(m1, i, j, 6);

  /* Cholesky decomposition */
  r = pom_dpotf2(6, a, 6, &e);
  if (r) {
    fprintf(stderr,"Your matrix is not invertible\n");
    return 1;
  }

  /* on inverse a triangulaire inferieure */
  for(i=0; i<6; i++) {
    a[i*6+i] = 1.0/a[i*6+i];
    for (j=i+1;j<6;j++) {
      sum = 0.0;
      for(k=i;k<j;k++)
	sum = sum - a[j*6+k]*a[k*6+i];
      a[j*6+i] = sum / a[j*6+j];
    }
  }

  /* btb dans m2 */
  for (i=0; i<6; i++) for(j=0; j<=i; j++) {
    sum = 0.0;
    for (k=(i>j?i:j); k<6; k++)
      sum += a[k*6+i]*a[k*6+j];
    *m2++ = sum;
  }

  return 0;
}


/* --- pomMatrixOfEuler ---------------------------------------------- */

/* Compute homogeneous matrix from Euler parameters. */

/* Pourquoi tant de haine? T3D sait faire a  ... */

void
pomMatrixOfEuler(double *euler, double *mat)
{
   double ca, sa, cb, sb, cg, sg;

   ca = cos(euler[0]); sa = sin(euler[0]);
   cb = cos(euler[1]); sb = sin(euler[1]);
   cg = cos(euler[2]); sg = sin(euler[2]);

   mat[0]  = ca*cb;
   mat[1]  = ca*sb*sg - sa*cg;
   mat[2]  = ca*sb*cg + sa*sg;
   mat[3]  = euler[3];

   mat[4]  = sa*cb;
   mat[5]  = sa*sb*sg + ca*cg;
   mat[6]  = sa*sb*cg - ca*sg;
   mat[7]  = euler[4];

   mat[8]  = -sb;
   mat[9]  = cb*sg;
   mat[10] = cb*cg;
   mat[11] = euler[5];

   mat[12] = 0.0;
   mat[13] = 0.0;
   mat[14] = 0.0;
   mat[15] = 1.0;
}


/*
 * --- pomDerivativeOfEuler -------------------------------------------
 *
 * Compute derivative rotation matrices of Euler angles
 */

void
pomDerivativeOfEuler(double *euler, double *dRa, double *dRb, double *dRg)
{
   double sa, ca, cb, sb, cg, sg;

   ca = cos(euler[0]); sa = sin(euler[0]);
   cb = cos(euler[1]); sb = sin(euler[1]);
   cg = cos(euler[2]); sg = sin(euler[2]);

   /* derivative /euler[0] */
   dRa[0] = -sa*cb;
   dRa[1] = -sa*sb*sg - ca*cg;
   dRa[2] = -sa*sb*cg + ca*sg;
   dRa[3] = ca*cb;
   dRa[4] = ca*sb*sg - sa*cg;
   dRa[5] = ca*sb*cg + sa*sg;
   dRa[6] = 0.0;
   dRa[7] = 0.0;
   dRa[8] = 0.0;

   /* derivative /euler[1] */
   dRb[0] = - ca*sb;
   dRb[1] = ca*cb*sg;
   dRb[2] = ca*cb*cg;
   dRb[3] = -sa*sb;
   dRb[4] = sa*cb*sg;
   dRb[5] = sa*cb*cg;
   dRb[6] = -cb;
   dRb[7] = -sb*sg;
   dRb[8] = -sb*cg;

   /* derivative /euler[2] */
   dRg[0] = 0.0;
   dRg[1] = ca*sb*cg + sa*sg;
   dRg[2] = -ca*sb*sg + sa*cg;
   dRg[3] = 0.0;
   dRg[4] = sa*sb*cg - ca*sg;
   dRg[5] = -sa*sb*sg - ca*cg;
   dRg[6] = 0.0;
   dRg[7] = cb*cg;
   dRg[8] = -cb*sg;
}
