/*	$Id: meset.c 2007/08/24 21:17:35 mallet $ */

/*
 * Copyright (c) 2003-2007 LAAS/CNRS                  --  Fri Aug 18 2000
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
static const char id[] = "$Id: meset.c 2007/08/24 21:17:35 mallet $";

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "libpom/lib.h"
#include "pomInt.h"



static int	pomMESetInit(PomMESet *set);

/**
 * pomMESetInit ---------------------------------------------------------
 *
 */

static int
pomMESetInit(PomMESet *set)
{
   int i, j;

   set->currentTime = 0;
   set->fusionIndex = 0;
   set->packets = POM_ME_PACKET_SIZE;
   set->length = POM_LE_PACKET_SIZE;

   set->refMe = -1;
   set->originToZero = t3dUMatrixIdentity;
   set->originToZero.flags = T3D_ALLOW_CONVERSION;

   set->log = NULL;

   set->me = calloc(set->packets,sizeof(PomMEHistory));
   if (set->me == NULL) {
      set->packets = 0;
      set->length = 0;
      errno = ENOMEM;
      return -1;
   }
   for(i=0; i<set->packets; i++) {
      set->me[i].p = calloc(set->length,sizeof(PomMEList *));
      if (set->me[i].p == NULL) {
	 set->packets = 0;
	 set->length = 0;
	 for(j=0;j<i;j++) free(set->me[j].p);
	 free(set->me);
      }
   }

   /** the first ME is the fused position */
   set->nmes = 1;
   set->me[0].fuse = 0;
   set->me[0].name = NULL; /** no need to define a name for this one */
   for(i=0;i<set->length;i++) set->me[0].p[i] = NULL;

   set->me[0].baseToOrigin = t3dUMatrixIdentity;
   set->me[0].baseToOrigin.flags = T3D_ALLOW_CONVERSION;
   set->me[0].lastabs.mainToOrigin = t3dUMatrixIdentity;
   set->me[0].lastabs.mainToOrigin.flags = T3D_ALLOW_CONVERSION;

   return 0;
}



/**
 * pomMESetResizeN ------------------------------------------------------
 *
 * Augment the number of position estimators to "nmes", if needed.
 */

int
pomMESetResizeN(PomMESet *set, int nmes)
{
   int i;

   if (set->packets == 0) if (pomMESetInit(set)) return -1;
   if (set->packets > nmes) return 0;

   set->packets += POM_ME_PACKET_SIZE;

   set->me = realloc(set->me, set->packets * sizeof(PomMEHistory));
   if (set->me == NULL) {
      set->packets = 0;
      errno = ENOMEM;
      return -1;
   }

   for(i = set->packets - POM_ME_PACKET_SIZE; i<set->packets; i++) {
      set->me[i].p = calloc(set->length, sizeof(PomMEList *));
      if (set->me[i].p == NULL) {
	 errno = ENOMEM;
	 return -1;
      }
   }

   return 0;
}


/**
 * pomMESetResizeL ------------------------------------------------------
 *
 * Augment the length of history to reach "oldest" date, if needed.
 */

int
pomMESetResizeL(PomMESet *set, int oldest)
{
   int s, i, j;

   if (set->packets == 0) if (pomMESetInit(set)) return -1;
   /** The 'oldest' date is already present in the indices */
   if (set->currentTime - set->length < oldest) return 0;

   /** Else, s is the new desired length */
   s = set->length;
   while (set->currentTime - set->length >= oldest)
      set->length += POM_LE_PACKET_SIZE;
   fprintf(stderr,"Augmenting size from %d to %d to reach %d (time %d) \n",s,set->length,oldest,set->currentTime);
   /** augment lists length */
   for(i = 0; i<set->packets; i++) {
      set->me[i].p = realloc(set->me[i].p, set->length*sizeof(PomMEList *));
      if (set->me[i].p == NULL) {
	 errno = ENOMEM;
	 return -1;
      }
      for(j=s;j<set->length;j++) set->me[i].p[j] = NULL;
   }

   return 0;
}



/**
 *
 * pomMEClass
 *
 * return the class of "me"
 */

PomMEClass
pomMEClass(PomMESet *set, int me)
{
   int i;

   /** Looks for the first index after "index" such taht we have a valid pose for me */
   for(i=0; i<set->length; i++)
     if (set->me[me].p[i] != NULL) return set->me[me].p[i]->pos.class;

   return POM_ABSOLUTE_ME;
}


/**
 * pomMELastPos ---------------------------------------------------------
 *
 * return the index of the last position  of "me" or -1.
 */

int
pomMELastPos(PomMESet *set, int me, int index)
{
   int i;

   /** Looks for the first index from "index" such taht we have a valid pose for me */
   for(i=index; i<set->length; i++)
      if (set->me[me].p[i] != NULL) return i;

   return -1;
}


/**
 * pomDateIndex ---------------------------------------------------------
 *
 * return the index inside a PomMeSet that correspond to the date "date"
 */

int
pomDateIndex(PomMESet *set, int date)
{
   int index;

   index = set->currentTime - date;
   /** If  the date is in the future, there is a small problem */
   if (index < 0) return -2;

   /** If we need to resize */
   if (index >= set->length) {
#ifdef POM_DEBUG
     fprintf(stderr,"Size %d < Index %d \n ",set->length,index);
#endif
     return -1;
     /** If the resizing fails */
     if (pomMESetResizeL(set, date)) return -1;
     /** Could it get this point ??? */
     if (index >= set->length) return -1;
   }

   /** Returns the index in the poses array */
   return index;
}


/**
 * pomBestME ------------------------------------------------------------
 *
 * return the index of the best position for "date".
 */

int
pomBestME(PomMESet *set, int date)
{
   int i, index,n;
   double conf = 0.0;

   index = -1;

   if (date>=POM_LE_PACKET_SIZE) return -1;

   /** Looks for the greater confidence coefficient */
   for(i=1,n=0;i<set->nmes;i++) if (set->me[i].fuse)
      if (set->me[i].p[date] != NULL &&
	  set->me[i].p[date]->pos.confidence > conf) {
	 conf = set->me[i].p[date]->pos.confidence;
	 index = i;
	 n++;
      }
   return index;
}
